import base64
import json
import pathlib
from typing import List, Any

import requests

from nldcsc.http_apis.base_class.api_base_class import ApiBaseClass


class MailgunAPI(ApiBaseClass):
    def __init__(
        self,
        baseurl: str,
        api_path: str = None,
        proxies: dict = None,
        user_agent: str = "MailGun",
        api_key: str = "",
        **kwargs,
    ):
        super().__init__(
            baseurl=baseurl,
            api_path=api_path,
            proxies=proxies,
            user_agent=user_agent,
            **kwargs,
        )

        self.set_header_field("access-token", f"{api_key}")

    def ping(self) -> bool:
        resource = ""
        try:
            data = self.call(self.methods.GET, resource=resource, timeout=5)
            if "ApplicationName" in data:
                return True
        except ConnectionError:
            return False

    def get_applications_list(self) -> dict[str, List[str] | str]:
        resource = "applications"
        return self.call(self.methods.GET, resource=resource)

    def get_application_template_list(self, application: str) -> dict[str, List[str]]:
        resource = f"applications/{application}"
        try:
            return self.call(self.methods.GET, resource=resource)
        except requests.exceptions.ConnectionError as err:
            return json.loads(str(err))

    def create_application(self, application: str) -> dict[str, str]:
        resource = f"applications/{application}"
        return self.call(self.methods.PUT, resource=resource)

    def delete_application(self, application: str) -> dict[str, str]:
        resource = f"applications/{application}"
        try:
            return self.call(self.methods.DELETE, resource=resource)
        except requests.exceptions.ConnectionError as err:
            return json.loads(str(err))

    def get_application_template(
        self, application: str, template_name: str
    ) -> dict[str, str]:
        resource = f"applications/{application}/template/{template_name}"
        try:
            return self.call(self.methods.GET, resource=resource)
        except requests.exceptions.ConnectionError as err:
            return json.loads(str(err))

    def download_application_template(
        self, application: str, template_name: str, output_path: pathlib.Path | str
    ) -> bool:
        template_data = self.get_application_template(application, template_name)

        if "errors" in template_data:
            raise RuntimeError(template_data["errors"])

        try:
            with open(output_path, "wb") as output_file:
                output_file.write(base64.b64decode(template_data["mjml_content"]))
            return True
        except KeyError:
            return False

    @staticmethod
    def fetch_template_from_location(location: str | pathlib.Path) -> str:
        try:
            with open(str(location), "r") as template:
                template_data = template.read()
            return template_data
        except FileNotFoundError:
            raise FileNotFoundError(f"Template file {location} not found")

    def create_application_template(
        self, application: str, template_name: str, template_path: pathlib.Path | str
    ) -> dict[str, str]:
        template_data = self.fetch_template_from_location(template_path)

        resource = f"applications/{application}/template/{template_name}"
        data = {
            "mjml_content": base64.b64encode(template_data.encode()).decode("utf-8"),
            "template": f"{template_name}",
        }
        try:
            return self.call(self.methods.POST, resource=resource, data=data)
        except requests.exceptions.ConnectionError as err:
            return json.loads(str(err))

    def update_application_template(
        self, application: str, template_name: str, template_path: pathlib.Path | str
    ) -> dict[str, str]:
        template_data = self.fetch_template_from_location(template_path)

        resource = f"applications/{application}/template/{template_name}"
        data = {
            "mjml_content": base64.b64encode(template_data.encode()).decode("utf-8"),
            "template": f"{template_name}",
        }
        try:
            return self.call(self.methods.PUT, resource=resource, data=data)
        except requests.exceptions.ConnectionError as err:
            return json.loads(str(err))

    def delete_application_template(
        self, application: str, template_name: str
    ) -> dict[str, str]:
        resource = f"applications/{application}/template/{template_name}"
        try:
            return self.call(self.methods.DELETE, resource=resource)
        except requests.exceptions.ConnectionError as err:
            return json.loads(str(err))

    def get_queue_length(self) -> dict[str, int]:
        resource = "queue"
        return self.call(self.methods.GET, resource=resource)

    def get_queue_schedule(self) -> dict[str, List[str] | int]:
        resource = "queue/schedule"
        return self.call(self.methods.GET, resource=resource)

    def get_queue_items(self) -> dict[str, List[str] | int]:
        resource = "queue/items"
        return self.call(self.methods.GET, resource=resource)

    def get_queue_item(self, item_name: str) -> dict[str, Any]:
        resource = f"queue/items/{item_name}"
        return self.call(self.methods.GET, resource=resource)

    def delete_queue_item(self, item_name: str) -> dict[str, Any]:
        resource = f"queue/items/{item_name}"
        return self.call(self.methods.DELETE, resource=resource)

    def send_mail_with_template(
        self, application: str, template_name: str, email_data: dict[str, Any]
    ) -> dict[str, List[str] | str]:
        """
        Check ../mail_objects.py for structure of email_data --> class Email (pydantic dependency needed if used)
        """
        resource = f"emails/{application}/template/{template_name}"
        return self.call(self.methods.POST, resource=resource, data=email_data)

    def send_batch_with_template(
        self, application: str, email_data: List[dict[str, Any]]
    ) -> dict[str, List[str] | str]:
        """
        Check ../mail_objects.py for structure of email_data --> class BatchEmail (pydantic dependency needed if used)
        """
        resource = f"emails/{application}/batch"
        return self.call(self.methods.POST, resource=resource, data=email_data)



import collections
import json
import logging
from dataclasses import dataclass, field
from typing import List, Optional

from dataclasses_json import config as json_config
from dataclasses_json import dataclass_json
from httpx import Response
from netaddr.ip import IPAddress
from nldcsc.generic.utils import exclude_optional_dict, reverse_from_named_tuple
from nldcsc.loggers.app_logger import AppLogger

from nldcsc.plugins.doh.doh_requester import DOHRequester, q_types

logging.setLoggerClass(AppLogger)

status_types = collections.namedtuple(
    "status_types",
    ["NOERROR", "FORMERR", "SERVFAIL", "NXDOMAIN", "NOTIMP", "REFUSED"],
)(0, 1, 2, 3, 4, 5)


def reverse_pointer_ip(ptr_address: str) -> str | None:
    """
    Method to reverse the reverse dns lookup address

    Args:
        ptr_address: reverse DNS lookup address

    Returns:
        original IP or None
    """
    ipv4_def = ".in-addr.arpa"
    ipv6_def = ".ip6.arpa"

    if ipv4_def in ptr_address:
        address = ptr_address.replace(ipv4_def, "")
        reverse_octets = address.split(".")[::-1]
        return ".".join(reverse_octets)

    if ipv6_def in ptr_address:
        address = ptr_address.replace(ipv6_def, "")
        reverse_string = address.replace(".", "")[::-1]
        full_ipv6_string = ":".join(
            reverse_string[i : i + 4] for i in range(0, len(reverse_string), 4)
        )
        ipv6_address = IPAddress(full_ipv6_string)
        return str(ipv6_address)

    return None


# noinspection PyClassHasNoInitInspection
@dataclass_json
@dataclass
class QuestionData:
    name: str
    type: int
    original_ip: Optional[str] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )

    @property
    def s_type(self) -> str:
        return reverse_from_named_tuple(q_types, self.type)


# noinspection PyClassHasNoInitInspection
@dataclass_json
@dataclass
class AnswerData:
    name: str
    type: int
    TTL: int
    data: str

    @property
    def s_type(self) -> str:
        return reverse_from_named_tuple(q_types, self.type)


# noinspection PyClassHasNoInitInspection
@dataclass_json
@dataclass
class Request:
    Status: int
    TC: bool
    RD: bool
    RA: bool
    AD: bool
    CD: bool
    Question: List[QuestionData]
    Answer: Optional[List[AnswerData]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    Authority: Optional[List[AnswerData]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )
    Comment: Optional[List[str]] = field(
        metadata=json_config(exclude=exclude_optional_dict), default=None
    )

    @property
    def s_status(self) -> str:
        return reverse_from_named_tuple(status_types, self.Status)

    @property
    def request_ok(self) -> bool:
        return self.Status == status_types.NOERROR

    @property
    def domain_exists(self) -> bool:
        return (
            not self.Status == status_types.NXDOMAIN
            and self.Status == status_types.NOERROR
        )

    @property
    def reverse_data_lookups(self) -> List[str]:
        if self.Status == status_types.NOERROR:
            if self.Answer is not None and len(self.Answer) > 0:
                return [
                    x.data for x in self.Answer if x.type in [q_types.A, q_types.AAAA]
                ]

        return []


class DohParser(object):
    def __init__(self):
        """
        Create new instance of DohParser
        """
        self.logger = logging.getLogger(self.__class__.__name__)
        self.doh_requester = DOHRequester("https://1.1.1.1")

    async def make_requests(
        self, workload: dict[int | str, List[str]]
    ) -> List[Response]:
        """
        Method to make DOH requests using the get_bulk_domains method of the DohRequester class

        Args:
            workload: a dictionary as dictated by get_bulk_domains method of the DohRequester class

        Returns:
            A list of `Response` objects
        """
        data = await self.doh_requester.get_bulk_domains(workload=workload)
        return data

    def parse_responses(self, responses: List[Response]) -> List[Request]:
        """
        Method to parse the responses returned from the get_bulk_domains method of the DohRequester class and parse
        these into `Request` objects

        Args:
            responses: A list of `Response` objects

        Returns:
            A list of `Request` objects
        """
        r_list = []
        for response in responses:
            if response.status_code == 200:
                try:
                    data = json.loads(response.content.decode("utf-8"))
                    if isinstance(data, dict):
                        if "Question" in data:
                            q_list = [
                                QuestionData(
                                    **x,
                                    original_ip=reverse_pointer_ip(x["name"]),
                                )
                                for x in data["Question"]
                            ]
                            data["Question"] = q_list
                        if "Answer" in data:
                            a_list = [AnswerData(**x) for x in data["Answer"]]
                            data["Answer"] = a_list
                        if "Authority" in data:
                            a_list = [AnswerData(**x) for x in data["Authority"]]
                            data["Authority"] = a_list

                        r_list.append(Request(**data))
                    else:
                        raise ValueError
                except TypeError:
                    self.logger.warning(
                        f"Response {response.content} yielded fields that could not be parsed..."
                    )
                    continue
                except UnicodeDecodeError:
                    self.logger.warning(f"Couldn't decode response: {response.content}")
                    continue
                except ValueError:
                    self.logger.warning(
                        f"Unexpected type in response: {response.content}"
                    )
                    continue
                except Exception:
                    self.logger.warning(
                        f"Uncaught error while parsing response: {response.content}"
                    )
                    continue
        return r_list

    async def request_and_parse(
        self, workload: dict[int | str, List[str]]
    ) -> dict[str, List[dict[str, str | int | List[dict[str, str | int]]]]]:
        query_data = await self.make_requests(workload=workload)

        # forward dns queries
        query_request_data = self.parse_responses(query_data)

        # reverse lookups
        reverse_lookups = []
        for response in query_request_data:
            reverse_lookups.extend(response.reverse_data_lookups)
        reverse_data = await self.make_requests(
            workload={q_types.PTR: list(set(reverse_lookups))}
        )

        reverse_request_data = self.parse_responses(reverse_data)

        ret_dict = {
            "forward_requests": [x.to_dict() for x in query_request_data],
            "reverse_requests": [x.to_dict() for x in reverse_request_data],
        }

        return ret_dict

    def __repr__(self):
        return "<< DohParser >>"



import os
import sys
import click
from importlib import util
import pkgutil

from nldcsc.plugins.sql_migrate import SqlMigrate


@click.group(invoke_without_command=True)
@click.option(
    "-v",
    "--version",
    is_flag=True,
    help="Show the current version of SqlMigrate and Alembic and exit",
)
@click.option(
    "-d",
    "--debug",
    is_flag=True,
    help="Set logging output to DEBUG",
)
@click.option("--config", help="Path to the sqlmigrate config", required=False)
@click.pass_context
def db(ctx, version, debug, config=None):
    """Perform database migrations."""

    if version:
        SqlMigrate.get_version(True)
        exit(0)

    if not config:
        click.echo("Please provide a config file!")
        exit(1)

    if debug:
        SqlMigrate.set_logger_debug_format()

    try:
        package = os.path.dirname(os.path.abspath(config))

        while "__init__.py" in os.listdir(package):
            package = os.path.dirname(package)
            config = os.path.relpath(os.path.abspath(config), package)

        os.chdir(package)

        sys.path.insert(0, package)
        spec = util.spec_from_file_location("sql_migrate_migrate_config", config)
        c = util.module_from_spec(spec)
        spec.loader.exec_module(c)

        sys.path.pop(0)
    except (AttributeError, FileNotFoundError, ModuleNotFoundError) as e:
        click.echo(f"Config file not found -> {e}, try running from the project root.")
        exit(1)

    if not hasattr(c, "db") or not hasattr(c, "metadata"):
        click.echo("Invalid config file, db and/or metadata does not exist!")
        exit(1)

    sql_migrate = SqlMigrate(c.db, c.metadata)

    ctx.ensure_object(SqlMigrate)
    ctx.obj = sql_migrate


@db.command()
@click.pass_context
def list_templates(ctx):
    """List available templates."""
    ctx.obj.list_templates(True)


@db.command()
@click.option(
    "-d",
    "--directory",
    default=None,
    help='Migration script directory (default is "migrations")',
)
@click.option(
    "-t",
    "--template",
    default=None,
    help='Repository template to use (default is "flask")',
)
@click.option(
    "--package",
    is_flag=True,
    help="Write empty __init__.py files to the environment and version locations",
)
@click.pass_context
def init(ctx, directory, template, package):
    """Creates a new migration repository."""
    ctx.obj.init(directory=directory, template=template, package=package)


@db.command()
@click.option("-m", "--message", default=None, help="Revision message")
@click.option(
    "--autogenerate",
    is_flag=True,
    help=(
        "Populate revision script with candidate migration operations, based on comparison of database to model"
    ),
)
@click.option(
    "--sql",
    is_flag=True,
    help="Don't emit SQL to database - dump to standard output instead",
)
@click.option(
    "--head",
    default="head",
    help="Specify head revision or <branchname>@head to base new revision on",
)
@click.option(
    "--splice",
    is_flag=True,
    help='Allow a non-head revision as the "head" to splice onto',
)
@click.option(
    "--branch-label",
    default=None,
    help="Specify a branch label to apply to the new revision",
)
@click.option(
    "--version-path",
    default=None,
    help="Specify specific path from config for version file",
)
@click.pass_context
def revision(
    ctx,
    message,
    autogenerate,
    sql,
    head,
    splice,
    branch_label,
    version_path,
):
    """Create a new revision file."""
    ctx.obj.revision(
        message=message,
        autogenerate=autogenerate,
        sql=sql,
        head=head,
        splice=splice,
        branch_label=branch_label,
        version_path=version_path,
    )


@db.command()
@click.option("-m", "--message", default=None, help="Revision message")
@click.option(
    "--sql",
    is_flag=True,
    help="Don't emit SQL to database - dump to standard output instead",
)
@click.option(
    "--head",
    default="head",
    help="Specify head revision or <branchname>@head to base new revision on",
)
@click.option(
    "--splice",
    is_flag=True,
    help='Allow a non-head revision as the "head" to splice onto',
)
@click.option(
    "--branch-label",
    default=None,
    help="Specify a branch label to apply to the new revision",
)
@click.option(
    "--version-path",
    default=None,
    help="Specify specific path from config for version file",
)
@click.pass_context
def migrate(ctx, message, sql, head, splice, branch_label, version_path):
    """Autogenerate a new revision file (Alias for 'revision --autogenerate')"""
    ctx.obj.migrate(
        message=message,
        sql=sql,
        head=head,
        splice=splice,
        branch_label=branch_label,
        version_path=version_path,
    )


@db.command()
@click.option(
    "--sql",
    is_flag=True,
    help="Don't emit SQL to database - dump to standard output instead",
)
@click.option(
    "--tag",
    default=None,
    help='Arbitrary "tag" name - can be used by custom env.py scripts',
)
@click.option(
    "-x",
    "--x-arg",
    multiple=True,
    help="Additional arguments consumed by custom env.py scripts",
)
@click.option(
    "--max-lookback-days",
    "-D",
    default=30,
    help="Amount of days revisions can be older than the current revision",
)
@click.option(
    "--sync",
    "-s",
    is_flag=True,
    help="Sync the old formatted migration files to the DB. Only available if no newly formatted migrations have been migrated.",
)
@click.option(
    "--fix-head",
    is_flag=True,
    help="Fix the head if it is not in sync with the current directory",
)
@click.argument("revision", default="head")
@click.pass_context
def upgrade(ctx, sql, tag, x_arg, revision, max_lookback_days, sync, fix_head):
    """Upgrade to a later version"""
    ctx.obj.upgrade(
        revision=revision,
        sql=sql,
        tag=tag,
        x_arg=x_arg,
        max_lookback_days=max_lookback_days,
        sync=sync,
        fix_head=fix_head,
    )


@db.command()
@click.option(
    "--max-lookback-days",
    "-D",
    default=30,
    help="Amount of days revisions can be older than the current revision",
)
@click.option(
    "--sync",
    "-s",
    is_flag=True,
    help="Sync the old formatted migration files to the DB. Only available if no newly formatted migrations have been migrated.",
)
@click.argument("revision", default=-1)
@click.pass_context
def check(ctx, revision, max_lookback_days, sync):
    """Upgrade to a later version"""
    if revision == -1:
        revision = None

    table_list = ctx.obj.check(
        revision=revision, max_lookback_days=max_lookback_days, sync=sync
    )

    if revision is None:
        click.echo("Missing migrations:")
    else:
        click.echo(f"Migration steps for {revision}:")

    if len(table_list) != 1:
        from tabulate import tabulate

        click.echo(tabulate(table_list, headers="firstrow", tablefmt="fancy_grid"))
    else:
        click.echo("None")


@db.command()
@click.option(
    "--sql",
    is_flag=True,
    help="Don't emit SQL to database - dump to standard output instead",
)
@click.option(
    "--tag",
    default=None,
    help='Arbitrary "tag" name - can be used by custom env.py scripts',
)
@click.option(
    "-x",
    "--x-arg",
    multiple=True,
    help="Additional arguments consumed by custom env.py scripts",
)
@click.argument("revision", default="-1")
@click.pass_context
def downgrade(ctx, sql, tag, x_arg, revision):
    """Downgrade to a earlier version"""
    ctx.obj.downgrade(revision=revision, sql=sql, tag=tag, x_arg=x_arg)


@db.command()
@click.pass_context
def drop(ctx):
    """Drop all tables in the database and clear migrations directory"""
    ctx.obj.drop()


@db.command()
@click.pass_context
def history(ctx):
    """Request the current migration history of the database"""
    table_list = ctx.obj.history()

    from tabulate import tabulate

    click.echo(tabulate(table_list, headers="firstrow", tablefmt="fancy_grid"))



from alembic.operations import Operations, MigrateOperation, BatchOperations
from alembic.operations import ops


def get_shell_obj(
    name: str, attr: str, cb: callable, reverse_obj: object = None, batch: bool = False
):
    if batch and attr.startswith("batch_"):
        batch_alt = attr
        attr = attr.removeprefix("batch_")
    elif batch:
        batch_alt = f"batch_{attr}"

    class shell_base(MigrateOperation):
        @classmethod
        def _classmethod(cls, operations, *args, **kwargs):
            cb(cls, operations, *args, **kwargs)

    def __init__(self, *args, **kwargs): ...

    def reverse(*args, **kwargs):
        if reverse_obj is not None:
            return reverse_obj(*args, **kwargs)

    dyn_class = {
        "__init__": __init__,
        attr: shell_base._classmethod,
        "reverse": reverse,
    }

    if batch:
        dyn_class[batch_alt] = shell_base._classmethod

    shell = type(name, (shell_base,), dyn_class)

    if batch:
        return BatchOperations.register_operation(attr, batch_alt)(shell)
    return Operations.register_operation(attr)(shell)


def create_shell(skip: set = None, cb: callable = lambda: ...):
    shell_objects = {}

    if skip is None:
        skip = set()

    for cls in [Operations, BatchOperations]:
        for attr in dir(cls):
            if attr in skip:
                # skip list
                continue

            op_name = "".join(p.capitalize() for p in attr.split("_")) + "Op"
            batch = False

            if cls == BatchOperations:
                batch = True

            if op_name.startswith("Batch"):
                continue

            if not hasattr(ops, op_name):
                continue

            shell_objects[attr] = get_shell_obj(f"{attr}_class", attr, cb, batch=batch)

    shell_objects["execute"] = get_shell_obj(
        "execute_class", "execute", cb, batch=False
    )
    shell_objects["execute"] = get_shell_obj("execute_class", "execute", cb, batch=True)

    return shell_objects



import logging

from typing import List
from nldcsc.http_apis.base_class.api_base_class import ApiBaseClass
from nldcsc.loggers.app_logger import AppLogger
from nldcsc.plugins.whoisxml.objects.whois import (
    WhoisRecord,
    WhoisReverseNSRecord,
    WhoisReverseMXRecord,
    WhoisReverseIPRecord,
)


logging.setLoggerClass(AppLogger)


class WhoisXMLAPI(ApiBaseClass):
    def __init__(
        self,
        api_key: str,
        baseurl: str = "https://www.whoisxmlapi.com",
        api_path: str = "",
        proxies: dict = None,
        user_agent: str = "Certex",
        **kwargs,
    ):
        super().__init__(
            baseurl=baseurl,
            api_path=api_path,
            proxies=proxies,
            user_agent=user_agent,
            **kwargs,
        )

        self.set_header_field("Authorization", f"Bearer {api_key}")

    def get_whois_record(
        self, domain_name: str, get_obj: bool = False
    ) -> dict | WhoisRecord:
        """
        https://whois.whoisxmlapi.com/documentation/making-requests

        Returns the whois record for the given domain name

        {
            "createdDate": "1997-09-15T00:00:00-0700",
            "updatedDate": "2018-02-21T10:45:07-0800",
            "expiresDate": "2020-09-13T21:00:00-0700",
            "registrant": {
              "organization": "Google LLC",
              "state": "CA",
              "country": "UNITED STATES",
              "countryCode": "US",
              "rawText": "Registrant Organization: Google LLC [...]"
            },
            "administrativeContact": {
              "organization": "Google LLC",
              "state": "CA",
              "country": "UNITED STATES",
              "countryCode": "US",
              "rawText": "Admin Organization: Google LLC [...]"
            },
            "technicalContact": {
              "organization": "Google LLC",
              "state": "CA",
              "country": "UNITED STATES",
              "countryCode": "US",
              "rawText": "Tech Organization: Google LLC [...]"
            },
            "domainName": "google.com",
            "nameServers": {
              "rawText": "ns2.google.com ns3.google.com ns1.google.com ns4.google.com",
              "hostNames": [
                "ns2.google.com",
                "ns3.google.com",
                "ns1.google.com",
                "ns4.google.com"
              ],
              "ips": []
            },
            "status": "clientUpdateProhibited [...]",
            "rawText": "Domain Name: google.com [...]",
            "parseCode": 3579,
            "header": "",
            "strippedText": "Domain Name: google.com [...]",
            "footer": "",
            "audit": {
              "createdDate": "2018-10-23 15:33:41.000 UTC",
              "updatedDate": "2018-10-23 15:33:41.000 UTC"
            },
            "customField1Name": "RegistrarContactEmail",
            "customField1Value": "abusecomplaints@markmonitor.com",
            "registrarName": "MarkMonitor, Inc.",
            "registrarIANAID": "292",
            "createdDateNormalized": "1997-09-15 07:00:00 UTC",
            "updatedDateNormalized": "2018-02-21 18:45:07 UTC",
            "expiresDateNormalized": "2020-09-14 04:00:00 UTC",
            "customField2Name": "RegistrarContactPhone",
            "customField3Name": "RegistrarURL",
            "customField2Value": "+1.2083895740",
            "customField3Value": "http://www.markmonitor.com",
            "registryData": {
              "createdDate": "1997-09-15T04:00:00Z",
              "updatedDate": "2018-02-21T18:36:40Z",
              "expiresDate": "2020-09-14T04:00:00Z",
              "domainName": "google.com",
              "nameServers": {
                "rawText": "NS1.GOOGLE.COM NS2.GOOGLE.COM NS3.GOOGLE.COM NS4.GOOGLE.COM",
                "hostNames": [
                  "NS1.GOOGLE.COM",
                  "NS2.GOOGLE.COM",
                  "NS3.GOOGLE.COM",
                  "NS4.GOOGLE.COM"
                ],
                "ips": []
              },
              "status": "clientDeleteProhibited [...]",
              "rawText": "Domain Name: GOOGLE.COM [...]",
              "parseCode": 251,
              "header": "",
              "strippedText": "Domain Name: GOOGLE.COM [...]",
              "footer": "",
              "audit": {
                "createdDate": "2018-10-23 15:33:40.000 UTC",
                "updatedDate": "2018-10-23 15:33:40.000 UTC"
              },
              "customField1Name": "RegistrarContactEmail",
              "customField1Value": "abusecomplaints@markmonitor.com",
              "registrarName": "MarkMonitor Inc.",
              "registrarIANAID": "292",
              "createdDateNormalized": "1997-09-15 04:00:00 UTC",
              "updatedDateNormalized": "2018-02-21 18:36:40 UTC",
              "expiresDateNormalized": "2020-09-14 04:00:00 UTC",
              "customField2Name": "RegistrarContactPhone",
              "customField3Name": "RegistrarURL",
              "customField2Value": "+1.2083895740",
              "customField3Value": "http://www.markmonitor.com",
              "whoisServer": "whois.markmonitor.com"
            },
            "domainAvailability": "UNAVAILABLE",
            "contactEmail": "abusecomplaints@markmonitor.com",
            "domainNameExt": ".com",
            "estimatedDomainAge": 7708,
            "ips": [
              "172.217.11.174"
            ]
        }
        """
        resource = (
            f"whoisserver/WhoisService?domainName={domain_name}&outputFormat=JSON&da=1"
        )
        response: dict = self.call(method=self.methods.GET, resource=resource)
        if get_obj:
            return WhoisRecord(**response["WhoisRecord"])
        else:
            return response


class WhoisReverseNSAPI(ApiBaseClass):
    def __init__(
        self,
        api_key: str,
        baseurl: str = "https://reverse-ns.whoisxmlapi.com",
        api_path: str = "api/v1",
        proxies: dict = None,
        user_agent: str = "Certex",
        **kwargs,
    ):
        super().__init__(
            baseurl=baseurl,
            api_path=api_path,
            proxies=proxies,
            user_agent=user_agent,
            **kwargs,
        )
        self.api_key = api_key

    def get_domains(
        self, ns_value: str, max_queries: int = None, get_obj: bool = False
    ) -> List[WhoisReverseNSRecord]:
        """


        Finds domain names associated with a nameserver using the WhoisXML Reverse NS API.
        Args:
            ns_value (str): The nameserver value to query for associated domain names.
            max_queries (int): The maximum number of queries to run. The API returns up to 300 results per query.
            get_obj (bool, optional): If True, returns a list of WhoisReverseNSRecord objects.
                    If False, returns a list of raw dictionary results. Defaults to False.
        Returns:
            List[WhoisReverseNSRecord] or List[dict]: A list of domain names associated with the given nameserver.
                    The format depends on the value of `get_obj`.
        Notes:
            - The API is paginated, and this method handles pagination automatically.
            - If the API response contains fewer than 300 results, it assumes there are no more pages.
            - Refer to the API documentation for more details:
                    https://reverse-ns.whoisxmlapi.com/api/documentation/making-requests

        """
        page = 1
        all_results = []
        query_count = 0
        while True:
            if max_queries and query_count >= max_queries:
                break
            resource = f"?apiKey={self.api_key}&ns={ns_value}&from={page}"
            response: dict = self.call(method=self.methods.GET, resource=resource)
            query_count += 1
            if "result" not in response or not response["result"]:
                break
            all_results.extend(response["result"])
            if len(response["result"]) < 300:
                break
            page = response["result"][-1]["name"]
        if get_obj:
            return [WhoisReverseNSRecord(**record) for record in all_results]
        else:
            return all_results


class WhoisReverseMXAPI(ApiBaseClass):
    def __init__(
        self,
        api_key: str,
        baseurl: str = "https://reverse-mx.whoisxmlapi.com",
        api_path: str = "api/v1",
        proxies: dict = None,
        user_agent: str = "Certex",
        **kwargs,
    ):
        super().__init__(
            baseurl=baseurl,
            api_path=api_path,
            proxies=proxies,
            user_agent=user_agent,
            **kwargs,
        )
        self.api_key = api_key

    def get_domains(
        self, mx_value: str, max_queries: int = None, get_obj: bool = False
    ) -> List[WhoisReverseMXRecord]:
        """


        Finds domain names associated with a mailserver using the WhoisXML Reverse MX API.
        Args:
            mx_value (str): The mailserver value to query for associated domain names.
            max_queries (int): The maximum number of queries to run. The API returns up to 300 results per query.
            get_obj (bool, optional): If True, returns a list of WhoisReverseMXRecord objects.
                    If False, returns a list of raw dictionary results. Defaults to False.
        Returns:
            List[WhoisReverseMXRecord] or List[dict]: A list of domain names associated with the given mailserver.
                    The format depends on the value of `get_obj`.
        Notes:
            - The API is paginated, and this method handles pagination automatically.
            - If the API response contains fewer than 300 results, it assumes there are no more pages.
            - Refer to the API documentation for more details:
                    https://reverse-mx.whoisxmlapi.com/api/documentation/making-requests

        """
        page = 1
        all_results = []
        query_count = 0
        while True:
            if max_queries and query_count >= max_queries:
                break
            resource = f"?apiKey={self.api_key}&mx={mx_value}&from={page}"
            response: dict = self.call(method=self.methods.GET, resource=resource)
            query_count += 1
            if "result" not in response or not response["result"]:
                break
            all_results.extend(response["result"])
            if len(response["result"]) < 300:
                break
            page = response["result"][-1]["name"]
        if get_obj:
            return [WhoisReverseMXRecord(**record) for record in all_results]
        else:
            return all_results


class WhoisReverseIPAPI(ApiBaseClass):
    def __init__(
        self,
        api_key: str,
        baseurl: str = "https://reverse-ip.whoisxmlapi.com",
        api_path: str = "api/v1",
        proxies: dict = None,
        user_agent: str = "Certex",
        **kwargs,
    ):
        super().__init__(
            baseurl=baseurl,
            api_path=api_path,
            proxies=proxies,
            user_agent=user_agent,
            **kwargs,
        )
        self.api_key = api_key

    def get_domains(
        self, ip_value: str, max_queries: int = None, get_obj: bool = False
    ) -> List[WhoisReverseIPRecord]:
        """


        Finds domain names associated with an ip using the WhoisXML Reverse IP API.
        Args:
            ip_value (str): The ip value to query for associated domain names.
            max_queries (int): The maximum number of queries to run. The API returns up to 300 results per query.
            get_obj (bool, optional): If True, returns a list of WhoisReverseIPRecord objects.
                    If False, returns a list of raw dictionary results. Defaults to False.
        Returns:
            List[WhoisReverseIPRecord] or List[dict]: A list of domain names associated with the given ip.
                    The format depends on the value of `get_obj`.
        Notes:
            - The API is paginated, and this method handles pagination automatically.
            - If the API response contains fewer than 300 results, it assumes there are no more pages.
            - Refer to the API documentation for more details:
                    https://reverse-ip.whoisxmlapi.com/api/documentation/making-requests
        """
        page = 1
        all_results = []
        query_count = 0
        while True:
            if max_queries and query_count >= max_queries:
                break
            resource = f"?apiKey={self.api_key}&ip={ip_value}&from={page}"
            response: dict = self.call(method=self.methods.GET, resource=resource)
            query_count += 1
            if "result" not in response or not response["result"]:
                break
            all_results.extend(response["result"])
            if len(response["result"]) < 300:
                break
            page = response["result"][-1]["name"]
        if get_obj:
            return [WhoisReverseIPRecord(**record) for record in all_results]
        else:
            return all_results



import logging

from typing import List
from nldcsc.loggers.app_logger import AppLogger
from nldcsc.plugins.hybrid_analysis.objects import HybridAnalysisHashRecord
import requests

logging.setLoggerClass(AppLogger)


class HybridAnalysisAPI:
    def __init__(
        self,
        api_key: str,
        baseurl: str = "https://www.hybrid-analysis.com",
        api_path: str = "api/v2",
        proxies: dict = None,
        user_agent: str = "Certex",
    ):
        self.url = f"{baseurl}/{api_path}"
        self.proxies = proxies
        self.headers = {}
        self.headers["api-key"] = api_key
        self.headers["User-Agent"] = user_agent
        self.session = requests.Session()
        self.session.headers = self.headers
        self.session.proxies = proxies

    def search_hash(
        self, hash: str, get_obj: bool = False
    ) -> dict | list[HybridAnalysisHashRecord]:
        data = {
            "hash": hash,
        }
        response: list[dict] = self.session.post(
            url=f"{self.url}/search/hash",
            data=data,
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                **self.headers,
            },
            proxies=self.proxies,
        ).json()

        if get_obj:
            records = []
            for record in response:
                records.append(HybridAnalysisHashRecord(**record))
            return records
        else:
            return response



import asyncio
from base64 import b64encode
from functools import wraps
from inspect import signature, Parameter
from json import dumps, loads

from redis import Redis
from redis.asyncio import Redis as RedisAsync


def compact_dump(value):
    return dumps(value, separators=(",", ":"), sort_keys=True)


def get_args(fn, args, kwargs):
    """
    This function parses the args and kwargs in the context of a function and creates unified
    dictionary of {<argument_name>: <value>}. This is useful
    because arguments can be passed as args or kwargs, and we want to make sure we cache
    them both the same. Otherwise there would be different caching for add(1, 2) and add(arg1=1, arg2=2)
    """
    arg_sig = signature(fn)
    standard_args = [
        param.name
        for param in arg_sig.parameters.values()
        if param.kind is param.POSITIONAL_OR_KEYWORD
        or param.kind is param.POSITIONAL_ONLY
    ]
    allowed_kwargs = {
        param.name
        for param in arg_sig.parameters.values()
        if param.kind is param.POSITIONAL_OR_KEYWORD or param.kind is param.KEYWORD_ONLY
    }
    variable_args = [
        param.name
        for param in arg_sig.parameters.values()
        if param.kind is param.VAR_POSITIONAL
    ]
    variable_kwargs = [
        param.name
        for param in arg_sig.parameters.values()
        if param.kind is param.VAR_KEYWORD
    ]
    parsed_args = {}

    if standard_args or variable_args:
        for index, arg in enumerate(args):
            try:
                parsed_args[standard_args[index]] = arg
            except IndexError:
                # then fallback to using the positional varargs name
                if variable_args:
                    vargs_name = variable_args[0]
                    if vargs_name not in parsed_args:
                        parsed_args[vargs_name] = []

                    parsed_args[vargs_name].append(arg)

    if kwargs:
        for key, value in kwargs.items():
            if key in allowed_kwargs:
                parsed_args[key] = value
            elif variable_kwargs:
                vkwargs_name = variable_kwargs[0]
                if vkwargs_name not in parsed_args:
                    parsed_args[vkwargs_name] = {}
                parsed_args[vkwargs_name][key] = value

    for param in arg_sig.parameters.values():
        if param.name not in parsed_args and param.default is not Parameter.empty:
            parsed_args[param.name] = param.default

    return parsed_args


def get_cache_lua_fn(client):
    if not hasattr(client, "_lua_cache_fn"):
        client._lua_cache_fn = client.register_script(
            """
            local ttl = tonumber(ARGV[2])
            local value
            if ttl > 0 then
              value = redis.call('SETEX', KEYS[1], ttl, ARGV[1])
            else
              value = redis.call('SET', KEYS[1], ARGV[1])
            end
            local limit = tonumber(ARGV[3])
            if limit > 0 then
              local time_parts = redis.call('TIME')
              local time = tonumber(time_parts[1] .. '.' .. time_parts[2])
              redis.call('ZADD', KEYS[2], time, KEYS[1])
              local count = tonumber(redis.call('ZCOUNT', KEYS[2], '-inf', '+inf'))
              local over = count - limit
              if over > 0 then
                local stale_keys_and_scores = redis.call('ZPOPMIN', KEYS[2], over)
                -- Remove the the scores and just leave the keys
                local stale_keys = {}
                for i = 1, #stale_keys_and_scores, 2 do
                  stale_keys[#stale_keys+1] = stale_keys_and_scores[i]
                end
                redis.call('ZREM', KEYS[2], unpack(stale_keys))
                redis.call('DEL', unpack(stale_keys))
              end
            end
            return value
            """
        )
    return client._lua_cache_fn


# Utility function to batch keys
def chunks(iterable, n):
    """Yield successive n-sized chunks from iterator."""
    _iterable = iter(iterable)
    while True:
        elements = []
        for _ in range(n):
            try:
                elements.append(next(_iterable))
            except StopIteration:
                break

        if not len(elements):
            break

        yield elements


async def async_chunks(iterable, n):
    """Yield successive n-sized chunks from iterator."""
    _iterable = aiter(iterable)
    while True:
        elements = []
        for _ in range(n):
            try:
                _e = await anext(_iterable)
                elements.append(_e)
            except StopAsyncIteration:
                break

        if not len(elements):
            break

        yield elements


class RedisCache:
    def __init__(
        self,
        redis_client: Redis | RedisAsync,
        prefix="rc",
        serializer=compact_dump,
        deserializer=loads,
        key_serializer=None,
        support_cluster=True,
        exception_handler=None,
        active: bool = True,
    ):
        self.client = redis_client
        self.prefix = prefix
        self.serializer = serializer
        self.deserializer = deserializer
        self.key_serializer = key_serializer
        self.exception_handler = exception_handler
        self.support_cluster = support_cluster
        self.active = active

    def cache(self, ttl=0, limit=0, namespace=None, exception_handler=None):
        return CacheDecorator(
            redis_client=self.client,
            prefix=self.prefix,
            serializer=self.serializer,
            deserializer=self.deserializer,
            key_serializer=self.key_serializer,
            ttl=ttl,
            limit=limit,
            namespace=namespace,
            support_cluster=self.support_cluster,
            exception_handler=exception_handler or self.exception_handler,
            active=self.active,
        )

    def set(self, name, value, **kwargs):
        if self.client and not isinstance(self.client, Redis):
            raise RuntimeError(
                "This method can only be used with a synchronous Redis client"
            )
        data = self.client.set(name, value, **kwargs)
        return data

    async def async_set(self, name, value, **kwargs):
        if self.client and not isinstance(self.client, RedisAsync):
            raise RuntimeError(
                "This method can only be used with an async Redis client"
            )
        data = await self.client.set(name, value, **kwargs)
        return data

    def get(self, name):
        if self.client and not isinstance(self.client, Redis):
            raise RuntimeError(
                "This method can only be used with a synchronous Redis client"
            )
        data = self.client.get(name)
        return data

    async def async_get(self, name):
        if self.client and not isinstance(self.client, RedisAsync):
            raise RuntimeError(
                "This method can only be used with an async Redis client"
            )
        data = await self.client.get(name)
        return data

    def mget(self, *fns_with_args):
        if self.client and not isinstance(self.client, Redis):
            raise RuntimeError(
                "This method can only be used with a synchronous Redis client"
            )
        keys = []
        for fn_and_args in fns_with_args:
            fn = fn_and_args["fn"]
            args = fn_and_args["args"] if "args" in fn_and_args else []
            kwargs = fn_and_args["kwargs"] if "kwargs" in fn_and_args else {}
            keys.append(fn.instance.get_key(args=args, kwargs=kwargs))

        results = self.client.mget(*keys)
        pipeline = self.client.pipeline()

        deserialized_results = []
        needs_pipeline = False
        for i, result in enumerate(results):
            if result is None:
                needs_pipeline = True

                fn_and_args = fns_with_args[i]
                fn = fn_and_args["fn"]
                args = fn_and_args["args"] if "args" in fn_and_args else []
                kwargs = fn_and_args["kwargs"] if "kwargs" in fn_and_args else {}
                result = fn.instance.original_fn(*args, **kwargs)
                result_serialized = self.serializer(result)
                get_cache_lua_fn(self.client)(
                    keys=[keys[i], fn.instance.keys_key],
                    args=[result_serialized, fn.instance.ttl, fn.instance.limit],
                    client=pipeline,
                )
            else:
                result = self.deserializer(result)
            deserialized_results.append(result)

        if needs_pipeline:
            pipeline.execute()
        return deserialized_results

    async def async_mget(self, *fns_with_args):
        if not isinstance(self.client, RedisAsync):
            raise RuntimeError(
                "This method can only be used with an async Redis client"
            )
        keys = []
        for fn_and_args in fns_with_args:
            fn = fn_and_args["fn"]
            args = fn_and_args["args"] if "args" in fn_and_args else []
            kwargs = fn_and_args["kwargs"] if "kwargs" in fn_and_args else {}
            keys.append(fn.instance.get_key(args=args, kwargs=kwargs))

        results = await self.client.mget(*keys)
        pipeline = self.client.pipeline()

        deserialized_results = []
        needs_pipeline = False
        for i, result in enumerate(results):
            if result is None:
                needs_pipeline = True

                fn_and_args = fns_with_args[i]
                fn = fn_and_args["fn"]
                args = fn_and_args["args"] if "args" in fn_and_args else []
                kwargs = fn_and_args["kwargs"] if "kwargs" in fn_and_args else {}
                if asyncio.iscoroutinefunction(fn.instance.original_fn):
                    result = await fn.instance.original_fn(*args, **kwargs)
                else:
                    result = await asyncio.to_thread(
                        fn.instance.original_fn, *args, **kwargs
                    )
                result_serialized = self.serializer(result)
                await get_cache_lua_fn(self.client)(
                    keys=[keys[i], fn.instance.keys_key],
                    args=[result_serialized, fn.instance.ttl, fn.instance.limit],
                    client=pipeline,
                )
            else:
                result = self.deserializer(result)
            deserialized_results.append(result)

        if needs_pipeline:
            await pipeline.execute()
        return deserialized_results


class CacheDecorator:
    def __init__(
        self,
        redis_client,
        prefix="rc",
        serializer=compact_dump,
        deserializer=loads,
        key_serializer=None,
        ttl=0,
        limit=0,
        namespace=None,
        support_cluster=True,
        exception_handler=None,
        active: bool = True,
    ):
        self.client = redis_client
        self.prefix = prefix
        self.serializer = serializer
        self.key_serializer = key_serializer
        self.deserializer = deserializer
        self.ttl = ttl
        self.limit = limit
        self.namespace = namespace
        self.exception_handler = exception_handler
        self.support_cluster = support_cluster
        self.keys_key = None
        self.original_fn = None
        self.active = active

    def get_full_prefix(self):
        if self.support_cluster:
            # Redis cluster requires keys operated in batch to be in the same key space. Redis cluster hashes the keys to
            # determine the key space. The braces specify which part of the key to hash (instead of the whole key).
            # See https://github.com/taylorhakes/python-redis-cache/issues/29  The `{prefix}:keys` and `{prefix}:args`
            # need to be in the same key space.
            return f"{{{self.prefix}:{self.namespace}}}"
        else:
            return f"{self.prefix}:{self.namespace}"

    def get_key(self, args, kwargs):
        normalized_args = get_args(self.original_fn, args, kwargs)

        if self.key_serializer:
            serialized_data = self.key_serializer(normalized_args)
        else:
            serialized_data = self.serializer(normalized_args)

        if isinstance(serialized_data, str):
            serialized_data = serialized_data.encode("utf-8")

        # Encode the value as base64 to avoid issues with {} and other special characters
        serialized_encoded_data = b64encode(serialized_data).decode("utf-8")

        return f"{self.get_full_prefix()}:{serialized_encoded_data}"

    def __call__(self, fn):
        self.namespace = self.namespace or f"{fn.__module__}.{fn.__qualname__}"
        self.keys_key = f"{self.get_full_prefix()}:keys"
        self.original_fn = fn

        if asyncio.iscoroutinefunction(fn):

            @wraps(fn)
            async def inner(*args, **kwargs):
                nonlocal self
                # Return the original function if we're not in active mode
                if not self.active:
                    return await fn(*args, **kwargs)
                key = self.get_key(args, kwargs)
                result = None

                exception_handled = False
                try:
                    if isinstance(self.client, Redis):
                        result = await asyncio.to_thread(self.client.get, key)
                    else:
                        result = await self.client.get(key)
                except Exception as e:
                    if self.exception_handler:
                        # This allows people to handle failures in cache lookups
                        exception_handled = True
                        if asyncio.iscoroutinefunction(self.exception_handler):
                            parsed_result = await self.exception_handler(
                                e, self.original_fn, args, kwargs
                            )
                        else:
                            parsed_result = await asyncio.to_thread(
                                self.exception_handler,
                                e,
                                self.original_fn,
                                args,
                                kwargs,
                            )
                if result:
                    parsed_result = self.deserializer(result)
                elif not exception_handled:
                    parsed_result = await fn(*args, **kwargs)
                    result_serialized = self.serializer(parsed_result)
                    if isinstance(self.client, Redis):
                        await asyncio.to_thread(
                            get_cache_lua_fn(self.client),
                            keys=[key, self.keys_key],
                            args=[result_serialized, self.ttl, self.limit],
                        )
                    else:
                        await get_cache_lua_fn(self.client)(
                            keys=[key, self.keys_key],
                            args=[result_serialized, self.ttl, self.limit],
                        )

                return parsed_result

            inner.invalidate = self.async_invalidate
            inner.invalidate_all = self.async_invalidate_all
        else:
            if self.client and not isinstance(self.client, Redis):
                raise RuntimeError(
                    "This method can only be used with a synchronous Redis client"
                )

            @wraps(fn)
            def inner(*args, **kwargs):
                nonlocal self
                # Return the original function if we're not in active mode
                if not self.active:
                    return fn(*args, **kwargs)
                key = self.get_key(args, kwargs)
                result = None

                exception_handled = False
                try:
                    result = self.client.get(key)
                except Exception as e:
                    if self.exception_handler:
                        # This allows people to handle failures in cache lookups
                        exception_handled = True
                        parsed_result = self.exception_handler(
                            e, self.original_fn, args, kwargs
                        )
                if result:
                    parsed_result = self.deserializer(result)
                elif not exception_handled:
                    parsed_result = fn(*args, **kwargs)
                    result_serialized = self.serializer(parsed_result)
                    get_cache_lua_fn(self.client)(
                        keys=[key, self.keys_key],
                        args=[result_serialized, self.ttl, self.limit],
                    )

                return parsed_result

            inner.invalidate = self.invalidate
            inner.invalidate_all = self.invalidate_all
        inner.get_full_prefix = self.get_full_prefix
        inner.instance = self
        return inner

    def invalidate(self, *args, **kwargs):
        key = self.get_key(args, kwargs)
        pipe = self.client.pipeline()
        pipe.delete(key)
        pipe.zrem(self.keys_key, key)
        pipe.execute()

    async def async_invalidate(self, *args, **kwargs):
        if isinstance(self.client, Redis):
            await asyncio.to_thread(self.invalidate, *args, **kwargs)
        else:
            key = self.get_key(args, kwargs)
            async with self.client.pipeline() as pipe:
                await pipe.delete(key)
                await pipe.zrem(self.keys_key, key)
                await pipe.execute()

    def invalidate_all(self, *args, **kwargs):
        chunks_gen = chunks(self.client.scan_iter(f"{self.get_full_prefix()}:*"), 500)
        for keys in chunks_gen:
            self.client.delete(*keys)

    async def async_invalidate_all(self, *args, **kwargs):
        if isinstance(self.client, Redis):
            await asyncio.to_thread(self.invalidate_all, *args, **kwargs)
        else:
            chunks_gen = async_chunks(
                self.client.scan_iter(f"{self.get_full_prefix()}:*"), 500
            )
            async for keys in chunks_gen:
                await self.client.delete(*keys)
